from llm_parametrizer.open_AI_models import GPT_MODEL
import json
import openai
import datetime
import os
import csv
import concurrent.futures
from dotenv import load_dotenv
from itertools import product
from datetime import datetime

class LLMParametrizer:
    def __init__(self, openai_key=None):
        self.all_parameters = []
        self.models = [GPT_MODEL.GPT_4o.value]
        self.roles = [{"role": "system", "content": "You are a helpful assistant."}]
        self.prompts = []
        self.messages = []
        self.temperatures = [1]
        self.seed = None # seed > -2,147,483,648 and seed < 2,147,483,647
        self.frequency_penalty = 0 # Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim.
        self.n = 1
        self.presence_penalty = 0 # Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics.
        self.response_format = { "type": "json_object" }
        self.top_p = 1 # Number between 0 and 1. An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10% probability mass are considered.
        self.tool_choice = "none"

    def initialize_OpenAI(self, openai_key=None):
        if openai_key:
            self.OPEN_AI_API_KEY = openai_key
        else:
            load_dotenv()
            self.OPEN_AI_API_KEY = os.getenv("OPEN_AI_API_KEY")
        self.client = openai.OpenAI(api_key=self.OPEN_AI_API_KEY)

    def add_prompts(self, *prompts):
        for prompt in prompts:
            self.prompts.append({"role": "user", "content": prompt})
        self.create_message()

    def add_roles(self, *roles):
        self.roles = []
        for role in roles:
            self.roles.append({"role": "system", "content": role})
        self.create_message()

    def create_message(self):
        combinations = product(self.roles, self.prompts)
        for role, prompt in combinations:
            self.messages.append([role, prompt])
        self.parametrize()

    def add_temperatures(self, *temperatures):
        self.temperatures = []
        for temperature in temperatures:
            self.temperatures.append(temperature)
        self.parametrize()

    def add_models(self, *models):
        self.models = []
        for model in models:
            self.models.append(model)
        self.parametrize()

    def add_tweaked_prompts(self, prompt_with_variables, values):
        keys = values.keys()
        values = values.values()

        combinations = list(product(*values))

        prompts = []
        for combination in combinations:
            temp_prompt = prompt_with_variables
            for key, value in zip(keys, combination):
                temp_prompt = temp_prompt.replace(f'{{{key}}}', value)
            prompts.append(temp_prompt)
        
        return prompts

    def return_parameters(self, return_raw=False):
        if return_raw:
            return self.all_parameters
        else:
            all_parameters = ""
            for parameters in self.all_parameters:
                prompt_system = parameters['messages'][0]['content']
                prompt_user = parameters['messages'][1]['content']
                temp = parameters['temperature']
                model = parameters['model']
                all_parameters += f"Prompt user: '{prompt_user}'\nPrompt system: '{prompt_system}'\nTemperature: {temp}\nModel: {model}\n\n"
            return all_parameters
    
    def show_parameters(self, show_raw=False):
        if show_raw:
            print(self.all_parameters)
        else:
            for parameters in self.all_parameters:
                prompt_system = parameters['messages'][0]['content']
                prompt_user = parameters['messages'][1]['content']
                temp = parameters['temperature']
                model = parameters['model']
                print(f"Prompt user: '{prompt_user}'\nPrompt system: '{prompt_system}'\nTemperature: {temp}\nModel: {model}\n")

    def chatGPT_API_call(self, parameters):

        completion = self.client.chat.completions.create(**parameters)

        response = completion.choices[0].message.content
        return response

    def run(self, return_raw=False, output_csv=False):
        assert self.prompts, "At least one prompt is required. use add_prompts() to add at least one prompt."
        date = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for parameters in self.all_parameters:
                future = executor.submit(self.chatGPT_API_call, parameters)
                futures.append((parameters, future))

            results = {} if return_raw else ""
            csv_data = []
            for parameters, future in futures:
                try:
                    response = future.result()
                    prompt_system = parameters['messages'][0]['content']
                    prompt_user = parameters['messages'][1]['content']
                    temp = parameters['temperature']
                    model = parameters['model']
                    if return_raw:
                        results[str(parameters)] = future.result()
                    else:
                        results += f"Prompt user: '{prompt_user}'\nPrompt system: '{prompt_system}'\nTemperature: {temp}\nModel: {model}\nDate: {date}\nResponse: {response}\n\n"

                    if output_csv:
                        csv_data.append([prompt_user, prompt_system, temp, model, date, response])

                except Exception as e:
                    if return_raw:
                        results[str(parameters)] = f"Error: {str(e)}"
                    else:
                        results += f"Error: {str(e)}\n"
            if output_csv:
                with open(f"results_{date}", mode='w', newline='') as file:
                    writer = csv.writer(file)
                    writer.writerow(["Prompt User", "Prompt System", "Temperature", "Model", "Time", "Response"])
                    writer.writerows(csv_data)

            return results

    def parametrize(self):
        combinations = product(self.messages, self.models, self.temperatures)
        self.all_parameters = [{'messages':messages, 'model': model, "temperature": temperature} for messages, model, temperature in combinations]




from enum import Enum

class GPT_MODEL(Enum):
    # GPT-4
    GPT_4_T = "gpt-4-turbo" # Points to a specific version of GPT-4 Turbo
    GPT_4_T_V = "gpt-4-turbo-2024-04-09" # SUpports vision and JSON mode
    GPT_4_T_P1 = "gpt-4-turbo-preview" # Points to a specific preview version of GPT-4 Turbo
    GPT_4_T_P2 = "gpt-4-0125-preview" # GPT-4 Turbo preview model intended to reduce cases of “laziness” where the model doesn’t complete a task.
    GPT_4_T_P3 = "gpt-4-1106-preview" # GPT-4 Turbo preview model featuring improved instruction following, JSON mode, reproducible outputs, parallel function calling, and more. 
    GPT_4_V = "gpt-4-vision-preview" # GPT-4 model with the ability to understand images, in addition to all other GPT-4 Turbo capabilities.
    GPT_4_V_P = "gpt-4-1106-vision-preview" # GPT-4 model with the ability to understand images, in addition to all other GPT-4 Turbo capabilities. 
    GPT_4 = "gpt-4" # Points to a specific version of GPT-4
    GPT_4_1 = "gpt-4-0613" # Snapshot of gpt-4 from June 13th 2023 with improved function calling support
    GPT_4_32 = "gpt-4-32k" # Points to a specific version of GPT-4-32k
    GPT_4_32_1 = "gpt-4-32k-0613" # Snapshot of gpt-4-32k from June 13th 2023 with improved function calling support. This model was never rolled out widely in favor of GPT-4 Turbo
    # GPT-4o
    GPT_4o = "gpt-4o" # Points to a specific version of GPT-4o
    GPT_4o_1 = "gpt-4o-2024-05-13" # gpt-4o v1
    # GPT-3.5
    GPT_3_5_T_1 = "gpt-3.5-turbo-0125" # New Updated GPT 3.5 Turbo
    GPT_3_5_T = "gpt-3.5-turbo" # Points to a specific version of GPT-3.5 Turbo
    GPT_3_5_T_P1 = "gpt-3.5-turbo-1106" # GPT-3.5 Turbo model with improved instruction following, JSON mode, reproducible outputs, parallel function calling.
    GPT_3_5_T_I = "gpt-3.5-turbo-instruct" # Similar capabilities as GPT-3 era models. Compatible with legacy Completions endpoint and not Chat Completions.
    GPT_3_5_T_16 = "gpt-3.5-turbo-16k" # Legacy Currently points to gpt-3.5-turbo-16k-0613.
    GPT_3_5_T_16_1 = "gpt-3.5-turbo-0613" # Legacy Snapshot of gpt-3.5-turbo from June 13th 2023. Will be deprecated on June 13, 2024.
    GPT_3_5_T_16_2 = "gpt-3.5-turbo-16k-0613" # Legacy Snapshot of gpt-3.5-16k-turbo from June 13th 2023. Will be deprecated on June 13, 2024.



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='llm_parametrizer',
    version='0.1.1',
    author ="Alejandro Camus",
    description="A library to parametrize multiple API calls for LLMs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['llm_parametrizer'],
    url="https://github.com/arjuna-dev/llm_parametrizer",
    install_requires=[
        'openai',
        'python-dotenv'
    ],
)